# Product:   Macal 2
# Author:    Marco Caspers
# Date:      16-10-2023
#
# String functions library

from typing import Any
from unidecode import unidecode



def StrLen(arg: str) -> int:
    """Implementation of len function"""
    return len(arg)



def StrLeft(arg: str, length: int) -> str:
    """Implementation of left function"""
    return arg[0:length]



def StrMid(arg: str, offset: int, length: int) ->str:
    """Implementation of mid function"""
    return arg[offset:(offset+length)]
    


def ToString(arg: Any) -> str:
    """Implementation of toString function"""
    if arg is True:
        return 'true'
    elif arg is False:
        return 'false'
    elif arg is None or arg == 'NIL':
        return 'nil'
    elif isinstance(arg, dict):
        vd = {k: 'nil' if v is None or v == 'NIL' else str(v).lower() if isinstance(v, bool) else v 
            for k, v in arg.items()}
        return f"{vd}"
    elif isinstance(arg, list):
        vl = ['nil' if v is None or v == 'NIL' else str(v).lower() if isinstance(v, bool) else v 
            for v in arg]
        return f"{vl}"
    return str(arg)        



def StrContains(needle: str, haystack: str) -> bool:
    """Implementation of strContains function"""
    if needle is None or haystack is None:
        return False
    return needle in haystack



def StrReplace(var: str, frm: str, wth: str) -> str:
    """Implementation of strReplace function"""
    if var is None:
        return None
    return var.replace(frm, wth)



def StartsWith(haystack: str, needle: str) -> bool:
    """Implementation of StartsWith function"""
    if haystack is None:
        return False
    return haystack.startswith(needle)



def RemoveNonAscii(text: str) -> str:
    """Implementation of RemoveNonAscii function"""
    if text is None:
        return None
    return unidecode(text)



def ReplaceEx(var: str, repl: Any, by: str) -> str:
    """Implementation of ReplaceEx function"""
    result = var
    if result is not None:
        for ch in repl:
            result = result.replace(ch, by)
    return result



def PadLeft(string: str, char: str, amount: int) -> Any:
    """Implementation of PadLeft function"""
    # this is counter intuitive, but the *just functions in python pad the character on the other 
    # end as what their name would imply.
    if string is None:
        string = ''
    if char is None:
        char = ' '
    if amount is None:
        amount = 0
    return string.rjust(amount, char)



def PadRight(string: str, char: str, amount: int) -> Any:
    """Implementation of PadRight function"""
    # this is counter intuitive, but the *just functions in python pad the character on the other 
    # end as what their name would imply.
    if string is None:
        string = ''
    if char is None:
        char = ' '
    if amount is None:
        amount = 0
    return string.ljust(amount, char)
    


def PadCenter(string: str, char: str, amount: int) -> Any:
    """Implementation of PadCenter function"""
    if string is None:
        string = ''
    if char is None:
        char = ' '
    if amount is None:
        amount = 0
    return string.center(amount, char)



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the expression ast node.

from __future__ import annotations
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node import AstNode

class AstNodeExpression(AstNode):
    def __init__(self, token: LexToken) -> AstNodeExpression:
        super().__init__()
        self._token: LexToken = token
        self.expr_type = AstNodetype.EXPRESSION

    @property
    def line(self) -> int:
        return self._token.lineno

    @property
    def column(self) -> int:
        return self._token.offset

    @property
    def token_value(self) -> str:
        return self._token.value

    def __str__(self) -> str:
        return f"AstNodeExpression({self._token})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}{self}")



# Product:   Macal 2
# Author:    Marco Caspers
# Date:      16-10-2023
#
# Syslog/logging extension for Macal 2

import sys
from typing import Any
import logging
import logging.config


class SysLog:
	def __init__(self):
		self.__address__ = 'localhost'
		self.__port__ = 514

		# REF, modified from:
		#   https://stackoverflow.com/a/19367225

		self.__config__ = {
			'version':1,
			'disable_existing_loggers': True,
			'formatters': {
				'verbose': {
					'format': '%(levelname)s %(module)s P%(process)d T%(thread)d %(message)s'
				},
				'simple': {
					'format': '%(levelname)s %(message)s'
				},
			},
			'handlers': {
				'stdout': {
					'class': 'logging.StreamHandler',
					'stream': sys.stdout,
					'formatter': 'simple',
				},
				'syslog0': {
					'class':     'logging.handlers.SysLogHandler',
					"address":   '/dev/log',
					'facility':   "local0",
					'formatter':  'simple',
				},
				'syslogR': {
					'class':     'logging.handlers.SysLogHandler',
					"address":   (self.__address__, self.__port__),
					'formatter':  'simple',
				},
			},
			'loggers': {
				'macal-syslog': {
					'handlers':  ['syslog0', 'stdout'],
					'level':     logging.DEBUG,
					'propagate': True,
				},
				'macal-syslog-remote': {
					'handlers':  ['syslogR'],
					'level':     logging.DEBUG,
					'propagate': False,
				},
			},
		}
		logging.config.dictConfig(self.__config__)
		self.handle   		 = None
		self.debug    		 = None
		self.info     		 = None
		self.warn     		 = None
		self.error    		 = None
		self.critical 		 = None
		self.syslog_enabled  = False
		self.remote_enabled  = False

	def SysLogInit(self, remote: bool):
		if (remote is True and self.remote_enabled) or (remote is False and self.syslog_enabled):
			raise Exception("Syslog is already initialized!")
		if remote is True:
			self.remote_enabled = True
			self.handle         = logging.getLogger('macal-syslog-remote')
			self.debug          = self.handle.debug
			self.info           = self.handle.info
			self.warn           = self.handle.warn
			self.error          = self.handle.error
			self.critical       = self.handle.critical
		else:
			self.syslog_enabled = True
			self.handle   		= logging.getLogger('macal-syslog')
			self.debug    		= self.handle.debug
			self.info     		= self.handle.info
			self.warn     		= self.handle.warn
			self.error    		= self.handle.error
			self.critical 		= self.handle.critical
		
	def SysLogSetAddress(self, address, port):
		if self.remote_enabled is True or self.syslog_enabled is True:
			raise Exception("Cannot change syslog configuration after initialization.")
		self.__address__ = address
		self.__port__ = port
		logging.config.dictConfig(self.__config__)

SysLogLocal = SysLog()
        
def Syslog(level: str, message: str) -> None:
    """Implementation of Syslog function"""
    if level == "debug" and SysLogLocal.debug is not None:
        SysLogLocal.debug(message)
    elif (level == "info" or level == "information") and SysLogLocal.info is not None:
        SysLogLocal.info(message)
    elif (level == "warn" or level == "warning") and SysLogLocal.warn is not None:
        SysLogLocal.warn(message)
    elif level == "error" and SysLogLocal.error is not None:
        SysLogLocal.error(message)
    elif level == "critical" and SysLogLocal.critical is not None:
        SysLogLocal.critical(message)
    else:
        raise Exception(f"Invalid syslog level given: {level}")

def SyslogInit(remote: bool) -> None:
    """Implementation of SysLog init function"""
    SysLogLocal.SysLogInit(remote)

def SyslogSetAddress(address: Any, port: Any) -> None:
    """Implementation of SysLog SetAddress function"""
    SysLogLocal.SysLogSetAddress(address, port)



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


"""
    This module contains a separate compiler for the select statement.
    The select statement is very complex and requires a lot of code.
    Therefore i have decided to put that compiler code in a separate module.

    The Select Statement is a complex statement that can be used to select data from a table, a function call,
    a record or an array.

    The Select Statement is composed of the following parts:

    SELECT [DISTINCT] <field> [AS <altfield>], <field> [AS <altfield>], ...
    FROM <table>
    [WHERE <expression>]
    [MERGE]
    INTO <variable>

    The SELECT clause is a comma separated list of fields to select.
    The fields can only be identifiers.
    It is possible to use the AS keyword to rename the field in the result set.
    It is possible to use the * wildcard to select all fields.

    Fields are not required to exist. Non existing fields will be automatically added, with a NIL value.

    The optional DISTINCT keyword is different from the distinct keyword in SQL as so far
    that it will ensure that only a single record returns, it will only return the first record that matches the
    where clause.  If the where clause is not present, it will return the first record from the from clause.

    The FROM clause is the table to select from.
    It can be a variable, a function call or a literal.
    The content is required to be an array of records, or a single record.
    If only a single record is presented, then the record will be
    automatically put inside an array as the only element of that array.

    The optional WHERE clause is an expression that must evaluate to true for the record to be selected.

    The optional MERGE clause is used to merge the selected record into a variable.  If the variable is an array,
    the record is appended to the array.  If the variable is a record, the record is merged into the variable.
    Note that when merging into a record, any existing fields in the record will be overwritten by the
    fields in the selected record if there is any overlap.

    The INTO clause is the variable to store the selected record or records into.
    It can only be a variable.

    Examples:

    select * from my_variable into my_variable2;
    select * from my_function() into my_variable;
    select a as b, d as e from my_variable where a == 3 into my_variable2;

    The bytecode VM is a stack machine.  The stack machine has a stack and a program counter.

    The stack is a list of values.
    The program counter is an integer that points to the next instruction to execute.

    The stack machine has a set of instructions that can be executed.
    Each instruction has an opcode and one or more optional operands.

    The stack machine has a set of registers that can be used to store values.
    The registers are named as follows:
    RAX, RBX, RCX, RDX, RSI, RDI, RSP, RBP, R8, R9, R10, R11, R12, R13, R14, R15, RIP, FLAGS
    Contrary to normal CPU registers these registers have a value and a type field.
    The value field contains the value of the register.
    The type field contains the type of the value, the AstNodetype enum is used for this.
    The stack is a list of stack entries. The stack entries also have a value and a type field.
    The value can be any valid Python value.


The ARRAY is represented in Python as a list.
The RECORD is represented in Python as a dictionary.
Otherwise the value types are identical to the Python types.

"""
from __future__ import annotations
from .ast_nodetype import AstNodetype
from .ast_node_select_statement import AstNodeSelectStatement
from .ast_node_select_field import AstNodeSelectField
from .ast_node_expression import AstNodeExpression
from .ast_node_binary_expression import AstNodeBinaryExpression
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_node_variable_expression import AstNodeVariableExpression
from .ast_node_unary_expression import AstNodeUnaryExpression
from .ast_node_indexed_variable_expression import AstNodeIndexedVariableExpression
import pyximport; pyximport.install()
from .cmacal_vm import MacalVm
from .macal_instructions import Opcode
from .bytecode_register import BytecodeRegister
from .compiler_scope import CompilerScope, CTVariable
from .macal_instruction_emitter import MacalInstructionEmitter
from .bytecode_debugger import ShowStack

# parent is the compiler, but we can't do that for reason of circular imports.
class MacalSelectCompiler:
    def __init__(self, parent) -> MacalSelectCompiler:
        self.do_raise: bool = parent.do_raise
        self.cs: MacalVm = parent.cs
        self.parent = parent
        self.emitter: MacalInstructionEmitter = parent.emitter



    def error(self, message: str) -> None:
        msg = f"SelectCompiler: {message}"
        if self.do_raise:
            raise Exception(msg)
        print(msg)
        import sys
        sys.exit(1)

    def compile_select(self, select: AstNodeSelectStatement, scope: CompilerScope) -> None:
        """
            This function compiles the given select statement into bytecode.
            The bytecode is returned as a list of tuples.
            Each tuple contains an opcode and a list of operands.
        """
        # SELECT FROM
        self.parent.compile_expression(select.From, scope)
        select_fields_lbl = scope.root.get_new_label("continue")
        
        # check if the result is empty.
        self.parent.emit_new_label("check_result_length", scope)
        self.emitter.MOVSR(self.cs.rax) # rax = stack[rsp-1]
        self.emitter.LEN(self.cs.rax) # array_register = len(rax)
        self.emitter.CMPRI(self.cs.rax, AstNodetype.INTEGER, 0) # if len == 0, jump to into_lbl
        self.parent.emit_jump(Opcode.JMPZ, select_fields_lbl, scope)

        # Check if the result is a record or an array.
        lbl = scope.root.get_new_label("from_continue")
        self.emitter.POP(self.cs.rax) # pop rax # rax has the value.
        self.emitter.CMPRTI(self.cs.rax, AstNodetype.RECORD)
        self.parent.emit_jump(Opcode.JMPNZ, lbl, scope)
        
        # It's a record, for easy processing we move it into an array.
        self.emitter.MOVRTOA(self.cs.rax) # rax = [rax]
        scope.root.set_label_address(lbl, self.cs.rip.value) # we jump here if it's an array or fall through to it if we have converted it to an array.
        self.emitter.PUSH(self.cs.rax) # push rax # rax has the array.
        
        if select.Where is not None:
            # Compile the where clause
            self.parent.emit_new_label("SELECT_WHERE", scope)
            self.compile_where(select, scope)

        # If there was no record to start with, we skipped to this.
        # If there was data to begin with we fall through to here via from and where.
        scope.root.set_label_address(select_fields_lbl, self.cs.rip.value) # into
        
        # Compile the select fields
        if not (len(select.Fields) == 1 and select.Fields[0].fieldname == '*'):
            self.emitter.POP(self.cs.rax) # rax = value
            self.compile_select_fields(select, scope)
            self.emitter.PUSH(self.cs.rax) # save value

        # Compile the merge clause, only if its required.
        if select.merge is True:
            self.parent.emit_new_label("SELECT_MERGE", scope)
            self.compile_merge(select, scope)
        
        # Compile the into clause
        self.compile_into(select, scope)        
     


    def compile_where(self, select: AstNodeSelectStatement, scope: CompilerScope) -> None:
        """
            This function compiles the where clause of the given select statement into bytecode.
            The bytecode is returned as a list of tuples.
            Each tuple contains an opcode and a list of operands.
        """
        # The core of this function is a loop that iterates over the array of records.
        # For each record it will evaluate the where clause.
        # If the where clause evaluates to true, the record will be added to the result array.
        # If the where clause evaluates to false, the record will be skipped.
        # The result array will be returned as the result of the where clause.
        # The result array will be a list of records, it will be empty if no records match the where clause.

        # Step 1: At the start of the where clause the array that was retrieved in the from clause is on the stack.
        # The rax register also has a copy of this.

        # We define labels for the start of the loop and the end of the loop.

        where_loop = scope.root.get_new_label("where_loop")
        where_end  = scope.root.get_new_label("where_end")

        # To know how much we need to loop for, we need to know the length of the array.
        # We will load the rcx register with the length. rcx is a special register that can be used in conjunction
        # with the djnz instruction.
        self.emitter.POP(self.cs.rax) # pop rax # rax has the array.
        self.emitter.LENR(self.cs.rcx, self.cs.rax) # rcx = len(rax) and can be used for the DJNZ instruction.

        # To safeguard we need to check if the array actually has any elements.
        # If it doesn't we can skip the loop.
        self.emitter.CMPRI(self.cs.rcx, AstNodetype.INTEGER, 0) # if rcx == 0, jump to where_end
        self.parent.emit_jump(Opcode.JMPZ, where_end, scope)

        # While RCX is used for iterating over the array, RBX is used for indexing into the array.
        # However RBX is a volatile register, so we acutally need another register to hold the index.
        index_register = self.cs.GetFreeRegister()
        self.emitter.XOR(index_register, index_register) # rbx = 0 and can be used for indexing.

        source_register = self.cs.GetFreeRegister() # we get a free register to hold the source array in.
        self.emitter.MOVRR(source_register, self.cs.rax) # source_register is now a "copy" of the array in rax.

        # We will request the VM for a free register to use for storing the resulting array of records
        result_register = self.cs.GetFreeRegister()
        self.emitter.MOVRI(result_register, AstNodetype.ARRAY, []) # result_register is now an empty array.

        # We will also need to request the VM for a free register to temporarily store the current record.
        # This is so we can use the current record in the equasions of the where clause.
        record_register = self.cs.GetFreeRegister()

        skip_append = scope.root.get_new_label("skip_append")

        scope.root.set_label_address(where_loop, self.cs.rip.value)  # this is where we return on each iteration of the loop.
        # At the start of the loop, we need the source array in the rax register and the index in the rbx register.
        self.emitter.MOVRR(self.cs.rax, source_register) # rax is now the array
        self.emitter.MOVRR(self.cs.rbx, index_register) # rbx is now the index
        self.emitter.INDEX() # rax = rax[rbx] rax is now the current record.
        self.emitter.MOVRR(record_register, self.cs.rax) # record_register is now a "copy" of the record in rax.

        self.compile_where_binary_expression(select.Where, record_register, scope) # we now execute the where clause.
        # Rax and rbx are destroyed in the process. The result will be on the stack. The result is either true or false.

        self.emitter.POP(self.cs.rax) # pop rax # rax has the result of the where clause.
        self.emitter.CMPRI(self.cs.rax, AstNodetype.BOOLEAN, True) # if rax == True we need to include the record in the result array.
        # If rax is false, we need to skip to the next record.
        self.parent.emit_jump(Opcode.JMPNZ, skip_append, scope) # if rax == False, jump to skip_1

        self.emitter.APPEND(result_register, record_register) # result_register[] = record_register
        # We now need to check if we need to continue to loop. If distinct is enabled, we can stop.
        self.emitter.MOVRI(self.cs.rax, AstNodetype.BOOLEAN, select.distinct) # rax = select.distinct
        self.emitter.CMPRI(self.cs.rax, AstNodetype.BOOLEAN, True) # if rax == True distinct is enabled and we can stop.
        self.parent.emit_jump(Opcode.JMPZ, where_end, scope) # if rax == True, jump to where_end
        scope.root.set_label_address(skip_append, self.cs.rip.value) # this is where we jump if we need to skip the record.
        self.emitter.INC(index_register) # index = index + 1
        self.parent.emit_jump(Opcode.DJNZ, where_loop, scope) # rcx = rcx -1, if rcx != 0 jump to where_loop
        scope.root.set_label_address(where_end, self.cs.rip.value) # this is where we jump if we need to skip the loop.

        # We now need to push the result array onto the stack.
        self.emitter.PUSH(result_register)
        # We can now release the registers that we used.
        self.cs.ReleaseRegister(source_register)
        self.cs.ReleaseRegister(index_register)
        self.cs.ReleaseRegister(result_register)
        self.cs.ReleaseRegister(record_register)



    def compile_select_fields(self, select: AstNodeSelectStatement, scope: CompilerScope) -> None:
        """
            This function compiles the code that will apply the field selection to the result of the from clause.
            (or the where clause if that exists)
        """
        self.parent.emit_new_label("fields", scope)
        # Compile the select fields
        # The fields are basically a filter on the result of the from clause or the where clause.
        # The result of the from clause or the where clause is on the stack.

        select_fields_end_label = scope.root.get_new_label("select_fields_end")
        select_null_record_label = scope.root.get_new_label("select_null_record")

        # We will request the VM for a free register to use for temporary storing array of records
        array_register = self.cs.GetFreeRegister()

        # skip this if there's no records.
        self.emitter.LENR(array_register, self.cs.rax) # array_register = len(rax)
        self.emitter.CMPRI(array_register, AstNodetype.INTEGER, 0) # if len == 0, jump to select_fields_continue
        self.parent.emit_jump(Opcode.JMPZ, select_null_record_label, scope) # jump to select_null_record_label

        self.emitter.MOVRR(array_register, self.cs.rax) # array_register is now a "copy" of the array in rax.

        # We need another register to store the result of the field selection in the array.
        result_array_register = self.cs.GetFreeRegister()
        self.emitter.MOVRI(result_array_register, AstNodetype.ARRAY, [])

        # we need yet another register to store the result for the field selection in the individual records.
        result_record_register = self.cs.GetFreeRegister()
        # We need another register to store the current record.
        current_record_register = self.cs.GetFreeRegister()
        # And another register to store the index register.
        index_register = self.cs.GetFreeRegister()
        # and another register to store the field value.
        field_value_register = self.cs.GetFreeRegister()

        # setup the index.
        self.emitter.XOR(self.cs.rbx, self.cs.rbx) # rbx = 0
        self.emitter.MOVRR(index_register, self.cs.rbx) # index_register = rbx (rbx is 0

        # setup the loop.
        array_loop_label = scope.root.get_new_label("allbl")
        array_continue_label = scope.root.get_new_label("aclbl")

        self.emitter.LENR(self.cs.rcx, self.cs.rax) # rcx = len(rax) and can be used for the DJNZ instruction.

        scope.root.set_label_address(array_loop_label, self.cs.rip.value)
        # the array loop starts here.
        # First we get RAX and RBX set up with the array and index.
        self.emitter.MOVRR(self.cs.rax, array_register) # rax is now the array again.
        self.emitter.MOVRR(self.cs.rbx, index_register) # rbx is now the index again.

        # We get the current record:
        self.emitter.INDEX() # rax = rax[rbx]
        # We check if it is a record:
        self.emitter.CMPRTI(self.cs.rax, AstNodetype.RECORD)
        self.emitter.ERRORNZ("Expected a record in the FROM clause of the SELECT statement.")
        # rax is now assured to be a record.
        self.emitter.MOVRR(current_record_register, self.cs.rax) # current_record_register is now a "copy" of the record in rax.
        # for each record that we process we need a new result record:
        self.emitter.MOVRI(result_record_register, AstNodetype.RECORD, {})

        for field in select.Fields: # we unroll this loop to prevent nested loops.
            self.compile_field(field, current_record_register, result_record_register, field_value_register, scope)
        # Append the result record to the result array.
        self.emitter.MOVRR(self.cs.rax, result_array_register) # rax is now the result array.
        self.emitter.APPEND(self.cs.rax, result_record_register) # rax[] = result_record_register

        # store the result array back into result_array_register.
        self.emitter.MOVRR(result_array_register, self.cs.rax) # result_array_register = rax

        # we're done with this record, so we increment the index.
        self.emitter.INC(index_register) # rbx = rbx + 1
        # and we check if we're done.
        # Only continue to loop if Distinct is False.
        # With Distinct True we only need the first matching record.
        if select.distinct is False:
            self.parent.emit_jump(Opcode.DJNZ, array_loop_label, scope) # rcx = rcx -1, if rcx != 0 jump to array_loop_label

        # if it falls through we're done.
        scope.root.set_label_address(array_continue_label, self.cs.rip.value)
        self.emitter.MOVRR(self.cs.rax, result_array_register) # rax is now the result array.
        self.cs.ReleaseRegister(array_register)
        self.cs.ReleaseRegister(result_array_register)
        self.cs.ReleaseRegister(result_record_register)
        self.cs.ReleaseRegister(current_record_register)
        self.cs.ReleaseRegister(index_register)
        self.cs.ReleaseRegister(field_value_register)
        # skip to the end.
        self.parent.emit_jump(Opcode.JMP, select_fields_end_label, scope) # jump to select_fields_end

        scope.root.set_label_address(select_null_record_label, self.cs.rip.value)
        # the record is empty but we do know the names of the fields that we need
        # so we create an empty record with those fields.
        self.emitter.MOVRI(self.cs.rax, AstNodetype.RECORD, {})
        self.emitter.MOVRI(self.cs.rcx, AstNodetype.NIL, AstNodetype.NIL)
        for field in select.Fields:
            self.emitter.MOVRI(self.cs.rbx, AstNodetype.STRING, field.altfieldname)
            self.emitter.INDEXR(self.cs.rcx)
            
        # this is the end of the select fields.
        scope.root.set_label_address(select_fields_end_label, self.cs.rip.value)



    def compile_field(self, field: AstNodeSelectField, record_register: BytecodeRegister,
                      result_record_register: BytecodeRegister, field_value_register: BytecodeRegister, scope: CompilerScope) -> None:
        new_record_field = self.cs.GetFreeRegister()
        self.emitter.MOVRI(new_record_field, AstNodetype.NIL, AstNodetype.NIL)
        has_fld_lbl = scope.root.get_new_label("has_fld")
        self.emitter.MOVRR(self.cs.rax, record_register) # rax is the current record
        self.emitter.MOVRI(self.cs.rbx, AstNodetype.STRING, field.fieldname) # rbx is the original name of the field
        self.emitter.HASFLDRR(self.cs.rax, self.cs.rbx)

        self.parent.emit_jump(Opcode.JMPZ, has_fld_lbl, scope)
        # The field does not exist in the record, so we need to create it.
        self.emitter.INDEXR(new_record_field)
        self.emitter.MOVRR(record_register, self.cs.rax)
        # We jump to this if the field did exist.
        scope.root.set_label_address(has_fld_lbl, self.cs.rip.value)
        self.emitter.INDEX() # rax = rax[rbx] rax is now the value of the field.
        self.emitter.MOVRI(self.cs.rbx, AstNodetype.STRING, field.altfieldname) # rbx is now the new field name.
        self.emitter.MOVRR(field_value_register, self.cs.rax) # field_value_register is now a "copy" of the field value in rax.
        self.emitter.MOVRR(self.cs.rax, result_record_register) # rax is now the result record.
        self.emitter.INDEXR(field_value_register) # # rax[rbx] = field_value_register so we add the current new field to the resulting record.
        self.emitter.MOVRR(result_record_register, self.cs.rax) # result_record_register is now a "copy" of the result record in rax.
        self.cs.ReleaseRegister(new_record_field)



    def compile_into(self, select: AstNodeSelectStatement, scope: CompilerScope) -> None:
        # Get the into variable if it exists, or create a new one, put the result on the stack.
        into: CTVariable = scope.find_variable(select.Into.name)       
        self.emitter.POP(self.cs.rax) # pop rax # rax has the value.
        reg = self.cs.GetFreeRegister()

        # process distinct to always return a record.
        lbl = scope.root.get_new_label("into_skip_1")
        self.emitter.LENR(reg, self.cs.rax)
        self.emitter.CMPRI(reg, AstNodetype.INTEGER, 0)
        self.parent.emit_jump(Opcode.JMPZ, lbl, scope) # if nothing here skip to lbl
        self.emitter.CMPRTI(self.cs.rax, AstNodetype.RECORD)
        self.parent.emit_jump(Opcode.JMPZ, lbl, scope) # if record here skip to lbl
        self.emitter.MOVRI(reg, AstNodetype.BOOLEAN, select.distinct)
        self.emitter.CMPRI(reg, AstNodetype.BOOLEAN, False)
        self.parent.emit_jump(Opcode.JMPZ, lbl, scope) # not distinct, so we skip this
        self.emitter.MOVATOR(self.cs.rax) # rax = rax[0] # we only need the first record.
        scope.set_label_address(lbl, self.cs.rip.value)

        # If the output is a single record, and it has only one field, and this is not * then the result is a single value.
        lbl2 = scope.root.get_new_label("into_skip_2")
        self.emitter.CMPRTI(self.cs.rax, AstNodetype.RECORD)
        self.parent.emit_jump(Opcode.JMPNZ, lbl2, scope) # if not a record here skip to lbl2
        self.emitter.MOVRI(reg, AstNodetype.BOOLEAN, select.merge)
        self.emitter.CMPRI(reg, AstNodetype.BOOLEAN, True)
        self.parent.emit_jump(Opcode.JMPZ, lbl2, scope) # if merge, skip to lbl2 because we already did do the merge
        # and we don't want to reduce the record to a single value if we're merging.
        self.emitter.MOVRI(reg, AstNodetype.INTEGER, len(select.Fields))
        self.emitter.CMPRI(reg, AstNodetype.INTEGER, 1)
        self.parent.emit_jump(Opcode.JMPNZ, lbl2, scope) # if not a single field, skip to lbl2
        self.emitter.MOVRI(reg, AstNodetype.STRING, select.Fields[0].fieldname)
        self.emitter.CMPRI(reg, AstNodetype.STRING, '*')
        self.parent.emit_jump(Opcode.JMPZ, lbl2, scope) # if it's "all" fields, then skip to lbl2
        self.emitter.MOVRR(self.cs.rbx, reg)
        self.emitter.INDEX() # RAX = RAX[RBX]
        scope.set_label_address(lbl2, self.cs.rip.value)
        self.cs.ReleaseRegister(reg)

        if into is None:
            self.parent.emit_new_label("into need a new var", scope)
            # it doesn't exist so we make one.
            reg = self.cs.GetFreeRegister()
            self.emitter.XOR(reg, reg)
            
            into = scope.define_variable(select.Into.name, self.cs.rsp.value)
            self.emitter.MOVRR(self.cs.rbx, self.cs.rsp) # rbx is now the absolute index of the new variable in the stack.
            self.emitter.PUSH(reg)
            self.cs.ReleaseRegister(reg)
        else:
            self.emitter.MOVRI(self.cs.rbx, AstNodetype.INTEGER, into.stack_offset)
        if select.Into.expr_type == AstNodetype.VARIABLE_EXPRESSION:
            self.emitter.STORIR(self.cs.rbx, self.cs.rax) # [rbx] = rax
            # we have now stored the result into the variable.
        elif select.Into.expr_type == AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            # rax = value to store
            # rbx = stack index of the variable
            self.walk_into_index(select.Into, scope)



    def walk_into_index(self
                   , expr: AstNodeExpression 
                   , scope: CompilerScope) -> None:
        reg = self.cs.GetFreeRegister()
        self.emitter.MOVRR(reg, self.cs.rax) # reg is now a copy of the value in rax.        
        self.emitter.LOADIR(self.cs.rax, self.cs.rbx) # rax = [rbx] # rax is now the value of the variable.
        self.emitter.PUSH(self.cs.rax) # push rax # the variable is now on top of the stack.
        n = len(expr.index)
        if n > 1:
            for i in range(n-1):
                self.parent.compile_expression(expr.index[i], scope)
                self.emitter.POP(self.cs.rbx)              # pop rbx
                self.emitter.POP(self.cs.rax)              # pop rax
                self.emitter.INDEX()        # rax = rax[rbx]
                self.emitter.PUSH(self.cs.rax)             # push rax
        self.parent.compile_expression(expr.index[n-1], scope)
        self.emitter.POP(self.cs.rbx)   # pop rbx
        self.emitter.POP(self.cs.rax)   # pop rax       
        self.emitter.INDEXR(reg)        # rax[rbx] = reg
        self.cs.ReleaseRegister(reg)
        


    def compile_merge(self, select: AstNodeSelectStatement, scope: CompilerScope) -> None:
        into: CTVariable = scope.find_variable(select.Into.name)
        if into is None:
            self.error(f"Variable not found ({select.Into.name}) at line: {select.Into.lineno}, column: {select.Into.offset}.")
        source_reg = self.cs.GetFreeRegister()
        data_reg = self.cs.GetFreeRegister()
        dest_reg = self.cs.GetFreeRegister()
        self.emitter.LOAD(source_reg, into.stack_offset)        # load the variable into source_reg.
        self.emitter.POP(data_reg)                          # pop data_reg with the data to merge.
        self.emitter.MOVRI(dest_reg, AstNodetype.ARRAY, [])     # setup the destination array.
        self.emitter.MERGE(dest_reg, source_reg, data_reg)      # merge the source and data into dest.
        self.emitter.PUSH(dest_reg)                         # push the result onto the stack.
        self.cs.ReleaseRegister(source_reg)
        self.cs.ReleaseRegister(data_reg)
        self.cs.ReleaseRegister(dest_reg)



# Where specific expression compilation.
    def compile_where_expression(self, expr: AstNodeExpression, record_register: BytecodeRegister, scope: CompilerScope):
        if expr.expr_type == AstNodetype.BINARY_EXPRESSION:
            self.compile_where_binary_expression(expr, record_register, scope)
        elif expr.expr_type == AstNodetype.LITERAL_EXPRESSION:
            self.compile_where_literal_expression(expr, scope)
        elif expr.expr_type == AstNodetype.FUNCTION_CALL_EXPRESSION:
            self.parent.compile_function_call_expression(expr, scope)
        elif expr.expr_type == AstNodetype.VARIABLE_EXPRESSION:
            self.compile_where_variable_expression(expr, record_register, scope)
        elif expr.expr_type == AstNodetype.UNARY_EXPRESSION:
            self.compile_where_unary_expression(expr, record_register, scope)
        elif expr.expr_type == AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            self.compile_where_indexed_variable_expression(expr, record_register, scope)
        else:
            self.error(f"Unexpected expression type {expr.expr_type.name} in WHERE clause.")



    def compile_where_binary_expression(self, where: AstNodeBinaryExpression,
                                        record_register: BytecodeRegister, scope: CompilerScope) -> None:
        # The were clause consists of binary expressions that will evaluate to a boolean value true or false.

        # The record to evaluate is in the rax register, a copy of that record is in the record_register.
        # The result of the where clause will be stored in the rax register.

        # The where clause is a binary expression, so we need to evaluate the left and right side of the expression.
        # The result of the left side of the expression will be stored in the rax register.
        # The result of the right side of the expression will be stored in the rbx register.
        # The result of the where clause will be stored in the rax register.

        # The left side of the expression is a fieldname in the record.
        # The operator is a comparison operator
        # The right side is a literal, a variable, a function call resulting in a literal, or another binary expression.
        # There may be multiple levels deep if the right hand side is a binary expression.
        # (This happens if and/or/not/xor is used in the where clause.)
        self.compile_where_expression(where.left, record_register, scope)
        # LHS is now on the stack.
        self.compile_where_expression(where.right, record_register, scope)
        # RHS is now on the stack.
        # We now need to compare the LHS and RHS using our operator.
        # The result of the comparison will be stored in the rax register.
        self.emitter.POP(self.cs.rbx)
        self.emitter.POP(self.cs.rax)
        self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
        # The result of the comparison is now in the FLAGS register.
        # Based on the flags and our operator we can set the rax register to True or False.
        if where.op.value == '==':
            self.emitter.SETZR(self.cs.rax)
        elif where.op.value == '!=':
            self.emitter.SETNZR(self.cs.rax)
        elif where.op.value == '<':
            self.emitter.SETLR(self.cs.rax)
        elif where.op.value == '<=':
            self.emitter.SETLER(self.cs.rax)
        elif where.op.value == '>':
            self.emitter.SETGR(self.cs.rax)
        elif where.op.value == '>=':
            self.emitter.SETGER(self.cs.rax)
        else:
            self.error(f"Illegal operator {where.op.value} in WHERE clause.")
        self.emitter.PUSH(self.cs.rax) # push result on register!



    def compile_where_literal_expression(self,
                                         expression: AstNodeLiteralExpression,
                                         scope: CompilerScope) -> None:
        # A literal expression means a literal value gets written to the stack.
        self.emitter.MOVRI(self.cs.rax, expression.value_type, expression.value)
        self.emitter.PUSH(self.cs.rax)



    def compile_where_unary_expression(self
                                       , expression: AstNodeUnaryExpression
                                       , record_register: BytecodeRegister
                                       , scope: CompilerScope) -> None:
        lbl = scope.root.get_new_label("unary_expression", scope)
        scope.root.set_label_address(lbl, self.cs.rip.value)
        self.compile_where_expression(expression.right, record_register, scope)
        self.emitter.POP(self.cs.rax) # pop rax
        if expression.op.token_type == AstNodetype.OPERATOR_SUBTRACTION:
            self.emitter.NEG(self.cs.rax)
        elif expression.op.token_type == AstNodetype.NOT_STATEMENT:
            self.emitter.NOT(self.cs.rax)
        elif expression.op.token_type == AstNodetype.BITWISE_NOT:
            self.error("Bitwise not is not yet implemented.")
        self.emitter.PUSH(self.cs.rax)



    def compile_where_variable_expression(self
                                          , expression: AstNodeVariableExpression
                                          , record_register: BytecodeRegister
                                          , scope: CompilerScope) -> None:
        lbl = scope.root.get_new_label("variable_where_expression")
        scope.root.set_label_address(lbl, self.cs.rip.value)
        var = scope.find_variable(expression.name)
        if var is None:
            # The variable should be a field in the record.
            # We need to push the value of the field onto the stack.
            self._record_field_to_stack(expression.name, record_register)
            return
        if not isinstance(var, CTVariable):
            self.error (f"Variable {expression.name} is not a variable at {expression.line}:{expression.column} scope({scope.name})")
        # variables live on the stack, the start of the stack for a scope is in the rbp register.
        # so we read the value from [rbp + var.offset] (our stack starts at 0 and counts up)
        # and push it to the stack.
        self.parent.var_to_stack(var)



    def compile_where_indexed_variable_expression(self
                                                  , expression: AstNodeIndexedVariableExpression
                                                  , record_register: BytecodeRegister
                                                  , scope: CompilerScope) -> None:
        lbl = scope.root.get_new_label("indexed_variable_expression")
        scope.root.set_label_address(lbl, self.cs.rip.value)
        var = scope.get_variable(expression.name)
        if var is None:
            # The variable is a field in the record.
            self._record_field_to_stack(expression.name, record_register, scope)
        else:
            # if it's just a var, get it.
            if not isinstance(var, CTVariable):
                self.error(f"Variable {expression.name} is not a variable at {expression.line}:{expression.column} scope({scope.name})")
            self.parent.var_to_stack(var) # store the var on the stack.
        self.walk_where_index(expression, record_register, scope)


######################### Shared Helper Functions #########################################

    def _record_field_to_stack(self, name: str, record_register: BytecodeRegister) -> None:
        # rax record , rax = rax[rbx], push rax
        self.emitter.MOVRR(self.cs.rax, record_register) # rax = record
        self.emitter.MOVRI(self.cs.rbx, AstNodetype.STRING, name) # rbx = name
        self.emitter.INDEX() # rax = rax[rbx]
        self.emitter.PUSH(self.cs.rax)



    def walk_where_index(self
                   , expression: AstNodeExpression 
                   , record_register: BytecodeRegister
                   , scope: CompilerScope) -> None:
        for expr in expression.index:                     # we walk each index part.
            self.compile_where_expression(expr, record_register, scope)    # the index value is on the stack now.
            self.emitter.POP(self.cs.rbx)             # pop rbx
            self.emitter.POP(self.cs.rax)             # pop rax
            self.emitter.INDEX()                          # rax = rax[rbx]
            self.emitter.PUSH(self.cs.rax)            # push rax

###########################################################################################################



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for the select statement.

from __future__ import annotations
from typing import Optional, List
from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node_select_field import AstNodeSelectField

class AstNodeSelectStatement(AstNodeStatement):
    def __init__(self, token: LexToken, fields: List[AstNodeSelectField], from_expr: AstNodeExpression, into: LexToken, merge: bool, distinct: bool, where: Optional[AstNodeExpression] = None) -> AstNodeSelectStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.SELECT_STATEMENT
        self.Fields: List[AstNodeSelectField] = fields
        self.From: AstNodeExpression = from_expr
        self.Where: Optional[AstNodeExpression] = where
        self.Into: LexToken = into
        self.merge: bool = merge
        self.distinct: bool = distinct

    def __str__(self) -> str:
        return f"SelectStatement({self.Fields}, {self.From}, {self.Where}, {self.Into})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}SelectStatement:")
        for field in self.Fields:
            field.tree(indent + '    ')
        self.From.tree(indent + '    ')
        if self.Where is not None:
            self.Where.tree(indent + '    ')
        print(f"{indent}    Into: {self.Into.name}")
        print(f"{indent}    Merge: {self.merge}")
        print(f"{indent}    Distinct: {self.distinct}")




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the parser for the Macal language. It is a recursive descent parser.

from __future__ import annotations
import sys
from typing import List, Optional, Tuple, Any

from .lex_token import LexToken
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .ast_node_binary_expression import AstNodeBinaryExpression
from .ast_node_unary_expression import AstNodeUnaryExpression
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_node_program import AstNodeProgram
from .ast_node_block import AstNodeBlock
from .ast_node_if_statement import AstNodeIfStatement
from .ast_node_elif_statement import AstNodeElifStatement
from .ast_node_else_statement import AstNodeElseStatement
from .ast_node_assignment import AstNodeAssignmentStatement
from .ast_node_indexed_variable_expression import AstNodeIndexedVariableExpression
from .ast_node_function_call_expression import AstNodeFunctionCallExpression
from .ast_node_variable_expression import AstNodeVariableExpression
from .ast_node_variable_function_call_expression import AstNodeVariableFunctionCallExpression
from .ast_node_print_statement import AstNodePrintStatement
from .ast_node_break_statement import AstNodeBreakStatement
from .ast_node_continue_statement import AstNodeContinueStatement
from .ast_node_return_statement import AstNodeReturnStatement
from .ast_node_halt_statement import AstNodeHaltStatement
from .ast_node_library_expression import AstNodeLibraryExpression
from .ast_node_include_statement import AstNodeIncludeStatement
from .ast_node_while_statement import AstNodeWhileStatement
from .ast_node_foreach_statement import AstNodeForeachStatement
from .ast_node_function_definition import AstNodeFunctionDefinition
from .ast_node_function_parameter import AstNodeFunctionParameter
from .ast_node_function_call_statement import AstNodeFunctionCallStatement
from .ast_node_select_statement import AstNodeSelectStatement
from .ast_node_select_field import AstNodeSelectField
from .ast_node_istype import AstNodeIsType
from .ast_node_type_statement import AstNodeTypeStatement
from .ast_node_default_statement import AstNodeDefaultStatement
from .ast_node_case_statement import AstNodeCaseStatement
from .ast_node_switch_statement import AstNodeSwitchStatement
from .conversion import convertToValue

class Parser:
    def __init__(self, tokens: Optional[List[LexToken]] = None, filename: Optional[str] = None) -> Parser:
        # remove the comments and the whitespace, we don't need that here.
        if tokens is None:
            tokens = []
        self.tokens: List[LexToken] = [token for token in tokens if token.token_type != AstNodetype.WHITESPACE and token.token_type != AstNodetype.COMMENT]
        self.filename: str = filename
        self.pos: int = 0
        self.current_token: Optional[LexToken] = None
        if len(self.tokens) > 0:
            self.current_token = self.tokens[self.pos]
        self.node: Optional[AstNodeProgram] = None
        self.do_raise: bool = True



    def error(self, message: str) -> None:
        msg = f"Parser error: {message} in {self.filename} at {self.current_token.lineno}, {self.current_token.offset}"
        if self.do_raise:
            raise Exception(msg)
        else:
            print(msg, file=sys.stderr)
            sys.exit(1)



    def strip_quotes(self, s: str) -> str:
        if (len(s) > 0) and (s[0] == '"' or s[0] == "'"):
            s = s[1:]
        if (len(s) > 0) and (s[-1] == '"' or s[-1] == "'"):
            s = s[:-1]
        return s



    def enforceLiteralType(self, type: AstNodetype, value: Any) -> Any:
        if type == AstNodetype.INTEGER and value != "integer":
            return int(value)
        elif type == AstNodetype.FLOAT and value != "float":
            return float(value)
        elif type in (AstNodetype.STRING,
                          AstNodetype.STRING_INTERPOLATION_STRING_PART,
                          AstNodetype.STRING_INTERPOLATION_END) and value != "string":
            sv = self.strip_quotes(value)
            return sv
        elif type == AstNodetype.BOOLEAN and value != "boolean":
            return bool(value=="true")
        elif type == AstNodetype.NIL and value != "nil":
            return AstNodetype.NIL
        return value



    def get_token(self, index: int) -> LexToken:
        token = self.tokens[index]
        token.value = self.enforceLiteralType(token.token_type, token.value)
        return token



    def advance(self) -> None:
        self.pos += 1
        if self.pos >= len(self.tokens):
            self.current_token = None
        else:
            self.current_token = self.get_token(self.pos)



    def peek(self) -> Optional[LexToken]:
        peek_pos = self.pos + 1
        if peek_pos >= len(self.tokens):
            return None
        else:
            return self.get_token(peek_pos)



    def expect(self, type: AstNodetype) -> Optional[LexToken]:
        token = self.current_token
        if token.token_type == type:
            self.advance()
            return token
        self.error(f"Expected {type} but got {token.token_type}")



    def parse(self) -> AstNodeExpression:
        self.node = self.program()
        return self.node
    


    def parse_interactive(self, tokens: List[LexToken]) -> AstNodeExpression:
        self.tokens = [token for token in tokens if token.token_type != AstNodetype.WHITESPACE and token.token_type != AstNodetype.COMMENT]
        self.pos = 0
        self.current_token = self.tokens[self.pos]
        self.node = self.program()
        return self.node



    def program(self) -> AstNodeExpression:
        node = AstNodeProgram(self.filename)
        self.block_statement_list(node)
        return node



    def block_statement_list(self, node: AstNodeBlock) -> AstNodeBlock:
        while self.current_token is not None and self.current_token.token_type != AstNodetype.RIGHT_BRACE and self.current_token.token_type != AstNodetype.EOF:
            if self.current_token.token_type != AstNodetype.EOF:
                node.statements.append(self.statement())
        return node



    def block(self) -> AstNodeBlock:
        node = AstNodeBlock(self.expect(AstNodetype.LEFT_BRACE))
        node = self.block_statement_list(node)
        self.advance() # skip right brace
        return node



    def statement(self) -> AstNodeExpression:
        if self.current_token.token_type == AstNodetype.IF_STATEMENT:
            return self.if_statement()
        if self.current_token.token_type == AstNodetype.WHILE_STATEMENT:
            return self.while_statement()
        if self.current_token.token_type == AstNodetype.FOREACH_STATEMENT:
            return self.foreach_statement()
        if self.current_token.token_type == AstNodetype.SELECT_STATEMENT:
            return self.select_statement()
        if self.current_token.token_type == AstNodetype.BREAK_STATEMENT:
            return self.break_statement()
        if self.current_token.token_type == AstNodetype.CONTINUE_STATEMENT:
            return self.continue_statement()
        if self.current_token.token_type == AstNodetype.RETURN_STATEMENT:
            return self.return_statement()
        if self.current_token.token_type == AstNodetype.IDENTIFIER:
            return self.assignment_statement()
        if self.current_token.token_type == AstNodetype.LEFT_BRACE:
            return self.block()
        if self.current_token.token_type == AstNodetype.PRINT_STATEMENT:
            return self.print_statement()
        if self.current_token.token_type == AstNodetype.HALT_STATEMENT:
            return self.halt_statement()
        if self.current_token.token_type == AstNodetype.INCLUDE_STATEMENT:
            return self.include_statement()
        if self.current_token.token_type == AstNodetype.SWITCH_STATEMENT:
            return self.switch_statement()
        if self.current_token.token_type == AstNodetype.CONST_STATEMENT:
            return self.const_statement()
        self.error(f"Invalid statement {self.current_token.token_type} {self.current_token.value}")



    def if_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.IF_STATEMENT)
        node = AstNodeIfStatement(token)
        node.condition = self.parse_expression()
        node.block = self.block()
        if self.current_token.token_type == AstNodetype.ELIF_STATEMENT:
            while self.current_token.token_type == AstNodetype.ELIF_STATEMENT:
                node.elif_block_list.append(self.elif_statement())
        if self.current_token.token_type == AstNodetype.ELSE_STATEMENT:
            node.else_block = self.else_statement()
        return node



    def elif_statement(self) -> AstNodeExpression:
        node = AstNodeElifStatement(self.current_token)
        self.advance()
        node.condition = self.parse_expression()
        node.statements = self.block()
        return node



    def else_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.advance()
        self.expect(AstNodetype.LEFT_BRACE)
        node = AstNodeElseStatement(token)
        self.block_statement_list(node)
        self.expect(AstNodetype.RIGHT_BRACE)
        return node



    def while_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.WHILE_STATEMENT)
        node = AstNodeWhileStatement(token)
        node.condition = self.parse_expression()
        node.block = self.block()
        return node



    def foreach_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.FOREACH_STATEMENT)
        return AstNodeForeachStatement(token, self.parse_expression(), self.block())



    def select_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.SELECT_STATEMENT)
        distinct = self.current_token.token_type == AstNodetype.DISTINCT_STATEMENT
        if distinct:
            self.advance()
        fields: List[AstNodeSelectField] = []
        while self.current_token.token_type == AstNodetype.IDENTIFIER:
            fieldname = self.current_token
            self.advance()
            alt = None
            if self.current_token.token_type == AstNodetype.AS_STATEMENT:
                self.advance()
                alt = self.current_token
                self.advance()
            fields.append(AstNodeSelectField(fieldname, alt))
            if self.current_token.token_type == AstNodetype.COMMA:
                self.advance()
        if len(fields) == 0 and self.current_token.token_type == AstNodetype.OPERATOR_MULTIPLICATION:
            fields.append(AstNodeSelectField(self.current_token))
            self.advance()
        self.expect(AstNodetype.FROM_STATEMENT)
        from_expr = self.parse_expression()
        where_expr: Optional[AstNodeExpression] = None
        if self.current_token.token_type == AstNodetype.WHERE_STATEMENT:
            self.advance()
            where_expr = self.parse_expression()
        merge = self.current_token.token_type == AstNodetype.MERGE_STATEMENT
        if merge:
            self.advance()
        self.expect(AstNodetype.INTO_STATEMENT)
        into = self.parse_expression()
        if into.expr_type != AstNodetype.VARIABLE_EXPRESSION and into.expr_type != AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            self.error(f"Invalid into expression type {into.expr_type} ({into.line}, {into.column})")
        self.expect(AstNodetype.SEMICOLON)
        return AstNodeSelectStatement(token, fields, from_expr, into, merge, distinct, where_expr)
        


    def break_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.BREAK_STATEMENT)
        self.expect(AstNodetype.SEMICOLON)
        return AstNodeBreakStatement(token)



    def continue_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.CONTINUE_STATEMENT)
        self.expect(AstNodetype.SEMICOLON)
        return AstNodeContinueStatement(token)



    def return_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.RETURN_STATEMENT)
        node = AstNodeReturnStatement(token)
        if self.current_token.token_type != AstNodetype.SEMICOLON:
            node.expr = self.parse_expression()
        self.expect(AstNodetype.SEMICOLON)
        return node



    def print_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.PRINT_STATEMENT)
        self.expect(AstNodetype.LEFT_PAREN)
        node = AstNodePrintStatement(token, self.expression_list())
        self.expect(AstNodetype.RIGHT_PAREN)
        self.expect(AstNodetype.SEMICOLON)
        return node



    def halt_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.HALT_STATEMENT)
        node = AstNodeHaltStatement(token)
        if self.current_token.token_type != AstNodetype.SEMICOLON:
            node.expr = self.parse_expression()
        self.expect(AstNodetype.SEMICOLON)
        return node



    def is_type_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.advance()
        var = self.parse_expression()
        if token.value == "IsString":
            return AstNodeIsType(token, var, AstNodetype.STRING)
        if token.value == "IsInt":
            return AstNodeIsType(token, var, AstNodetype.INTEGER)
        if token.value == "IsFloat":
            return AstNodeIsType(token, var, AstNodetype.FLOAT)
        if token.value == "IsBool":
            return AstNodeIsType(token, var, AstNodetype.BOOLEAN)
        if token.value == "IsRecord":
            return AstNodeIsType(token, var, AstNodetype.RECORD)
        if token.value == "IsArray":
            return AstNodeIsType(token, var, AstNodetype.ARRAY)
        if token.value == "IsFunction":
            return AstNodeIsType(token, var, AstNodetype.FUNCTION)
        if token.value == "IsNil":
            return AstNodeIsType(token, var, AstNodetype.NIL)



    def type_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.TYPE_STATEMENT)
        return AstNodeTypeStatement(token, self.parse_expression())



    def library_list(self) -> List[AstNodeExpression]:
        token = self.current_token
        self.expect(AstNodetype.IDENTIFIER)
        nodes = [AstNodeLibraryExpression(token)]
        while self.current_token.token_type == AstNodetype.COMMA:
            self.advance()
            token = self.current_token
            self.expect(AstNodetype.IDENTIFIER)
            nodes.append(AstNodeLibraryExpression(token))
        return nodes



    def include_statement(self) -> AstNodeExpression:
        token = self.current_token
        self.expect(AstNodetype.INCLUDE_STATEMENT)
        node = AstNodeIncludeStatement(token, self.library_list())
        self.expect(AstNodetype.SEMICOLON)
        return node



    def switch_statement(self) -> AstNodeExpression:
        token = self.expect(AstNodetype.SWITCH_STATEMENT)
        expr = self.parse_expression()
        if expr.expr_type not in [AstNodetype.VARIABLE_EXPRESSION, AstNodetype.INDEXED_VARIABLE_EXPRESSION,
                             AstNodetype.FUNCTION_CALL_EXPRESSION, AstNodetype.VARIABLE_FUNCTION_CALL_EXPRESSION]:
            # may not need line,col here, but just in case.
            self.error(f"Invalid switch expression type {expr.expr_type} ({expr.line}, {expr.column})")
        node = AstNodeSwitchStatement(token, expr)
        self.expect(AstNodetype.LEFT_BRACE)
        while self.current_token.token_type == AstNodetype.CASE_STATEMENT:
            node.cases.append(self.case_statement())
        if self.current_token.token_type == AstNodetype.DEFAULT_STATEMENT:
            node.default = self.default_statement()
        self.expect(AstNodetype.RIGHT_BRACE)
        return node



    def case_statement(self) -> AstNodeExpression:
        token = self.expect(AstNodetype.CASE_STATEMENT)
        expr = self.parse_expression()
        if expr.expr_type != AstNodetype.LITERAL_EXPRESSION:
            # may not need line,col here, but just in case.
            self.error(f"Invalid case expression type {expr.expr_type} ({expr.line}, {expr.column})")
        node = AstNodeCaseStatement(token, expr)
        self.expect(AstNodetype.COLON)
        node.block = self.block()
        return node



    def default_statement(self) -> AstNodeExpression:
        token = self.expect(AstNodetype.DEFAULT_STATEMENT)
        self.expect(AstNodetype.COLON)
        node = AstNodeDefaultStatement(token)
        node.block = self.block()
        return node



    def expression_list(self) -> List[AstNodeExpression]:
        nodes: List[AstNodeExpression] = []
        if self.current_token.token_type == AstNodetype.RIGHT_PAREN:
            return nodes
        nodes.append(self.parse_expression())
        while self.current_token.token_type == AstNodetype.COMMA:
            self.advance()
            nodes.append(self.parse_expression())
        return nodes



    def expression_array_element(self) -> Any:
        node = self.parse_expression()
        if (node.expr_type != AstNodetype.LITERAL_EXPRESSION):
            self.error(f"Invalid array element type {node.expr_type} (expected literal, array or record) ({node.token.lineno}, {node.token.offset})")
        vt = node.value_type
        if (vt == AstNodetype.STRING_INTERPOLATION_STRING_PART
            or vt == AstNodetype.STRING_INTERPOLATION_END):
                self.error(f"Invalid array element type {vt} (expected literal, array or record) ({node.token.lineno}, {node.token.offset})")
        if vt == AstNodetype.ARRAY or vt == AstNodetype.RECORD:
            return node.value
        return convertToValue(node._token)



    def expression_array(self) -> list:
        if self.current_token.token_type == AstNodetype.RIGHT_BRACKET:
            self.error(f"Invalid array element, expected literal, got ']'")
        array: list = []
        array.append(self.expression_array_element())
        while self.current_token.token_type == AstNodetype.COMMA:
            self.advance()                      # skip comma
            array.append(self.expression_array_element())
        return array



    def expression_record_element(self) -> Tuple[Any, Any]:
        key: AstNodeExpression = self.parse_expression()
        if key.expr_type != AstNodetype.LITERAL_EXPRESSION:
            self.error(f"Invalid record key type {key.expr_type} (expected literal) ({key.line}, {key.column})")
        self.expect(AstNodetype.COLON)
        value = self.parse_expression()
        if value.expr_type != AstNodetype.LITERAL_EXPRESSION:
            self.error(f"Invalid record value type {value.expr_type} (expected literal, array or record) ({value.line}, {value.column})")
        if (key.expr_type == AstNodetype.STRING_INTERPOLATION_STRING_PART
                or key.expr_type == AstNodetype.STRING_INTERPOLATION_END):
            self.error(f"Invalid record key type {key.expr_type} (expected literal) ({key.line}, {key.column})")
        if (value.expr_type == AstNodetype.LITERAL_EXPRESSION and (value.expr_type == AstNodetype.STRING_INTERPOLATION_STRING_PART
                    or value.expr_type == AstNodetype.STRING_INTERPOLATION_END)):
            self.error(f"Invalid record value type {value.expr_type} (expected literal, array or record) ({value.line}, {value.column})")
        return (key.value, value.value)



    def expression_record(self) -> dict:
        if self.current_token.token_type == AstNodetype.RIGHT_BRACE:
            self.error(f"Invalid record element, expected record key, got '}}' ")
        record: dict = {}
        (key, value) = self.expression_record_element()
        record[key] = value
        while self.current_token.token_type == AstNodetype.COMMA:
            self.advance()                      # skip comma
            (key, value) = self.expression_record_element()
            record[key] = value
        return record



    def expect_assign_operand(self) -> LexToken:
        token = self.current_token
        if (token.token_type == AstNodetype.OPERATOR_ASSIGNMENT
            or token.token_type == AstNodetype.OPERATOR_ASSIGNMENT_INC
            or token.token_type == AstNodetype.OPERATOR_ASSIGNMENT_DEC
            or token.token_type == AstNodetype.OPERATOR_ASSIGNMENT_MUL
            or token.token_type == AstNodetype.OPERATOR_ASSIGNMENT_DIV
            or token.token_type == AstNodetype.OPERATOR_ASSIGNMENT_POW
            or token.token_type == AstNodetype.OPERATOR_ASSIGNMENT_MOD
            or token.token_type == AstNodetype.OPERATOR_ARROW):
            self.advance()
            return token
        self.error(f"Invalid assignment operator {token.token_type} {token.value} ")



    def assignment_statement(self) -> AstNodeExpression:
        var_expr = self.parse_expression() # Get the variable as an expression.
        if var_expr.expr_type in [AstNodetype.FUNCTION_CALL_EXPRESSION,
                             AstNodetype.VARIABLE_FUNCTION_CALL_EXPRESSION]:
            self.expect(AstNodetype.SEMICOLON)
            return AstNodeFunctionCallStatement(var_expr._token, var_expr.args)
        if not var_expr.expr_type in [AstNodetype.VARIABLE_EXPRESSION,
                                 AstNodetype.INDEXED_VARIABLE_EXPRESSION]:
            self.error(f"Invalid assignment statement {var_expr.expr_type} {var_expr._token.value} ({var_expr._token.lineno}, {var_expr._token.offset})")
        if self.current_token.token_type == AstNodetype.LEFT_PAREN:
            self.expect(AstNodetype.LEFT_PAREN)
            args = self.expression_list()
            node = AstNodeFunctionCallStatement(var_expr, args)
            self.expect(AstNodetype.RIGHT_PAREN)
            self.expect(AstNodetype.SEMICOLON)
            return node
        if self.current_token.token_type == AstNodetype.NEW_ARRAY:
            self.advance()
            node = AstNodeAssignmentStatement(var_expr, self.expect_assign_operand(), self.parse_expression())
            node.append = True
            self.expect(AstNodetype.SEMICOLON)
            return node
        op = self.expect_assign_operand()
        if op.token_type == AstNodetype.OPERATOR_ARROW:
            return self.function_definition(var_expr, op)
        assng = AstNodeAssignmentStatement(var_expr, op, self.parse_expression())
        self.expect(AstNodetype.SEMICOLON)
        return assng



    def function_definition(self, var_expr: AstNodeExpression, op: LexToken) -> AstNodeFunctionDefinition:
        params = self.parse_parameter_list()
        return_type = AstNodetype.NIL
        if self.is_type():
            return_type = self.current_token.token_type
            self.advance()
        if self.current_token.token_type == AstNodetype.EXTERNAL_STATEMENT:
            self.advance()
            fndef = AstNodeFunctionDefinition(var_expr, op, params, return_type, None)
            fndef.external = True
            module = self.expect(AstNodetype.STRING)
            fndef.module = self.strip_quotes(module.value)
            self.expect(AstNodetype.COMMA)
            function = self.expect(AstNodetype.STRING)
            fndef.function = self.strip_quotes(function.value)
            self.expect(AstNodetype.SEMICOLON)
            return fndef
        return AstNodeFunctionDefinition(var_expr, op, params, return_type, self.block())



    def const_statement(self) -> AstNodeExpression:
        self.expect(AstNodetype.CONST_STATEMENT)
        var = self.peek()
        result = self.assignment_statement()
        result.const = True
        return result



    def is_type(self) -> bool:
        if self.current_token.token_type in [
            AstNodetype.INTEGER,
            AstNodetype.FLOAT,
            AstNodetype.STRING,
            AstNodetype.BOOLEAN,
            AstNodetype.ARRAY,
            AstNodetype.RECORD,
            AstNodetype.FUNCTION,
            AstNodetype.PARAMS,
            AstNodetype.VARIABLE]:
            return True
        return False



    def strToType(self, tstr: str) -> AstNodetype:
        if tstr == "integer": return AstNodetype.INTEGER
        if tstr == "float": return AstNodetype.FLOAT
        if tstr == "string": return AstNodetype.STRING
        if tstr == "bool": return AstNodetype.BOOLEAN
        if tstr == "array": return AstNodetype.ARRAY
        if tstr == "record": return AstNodetype.RECORD
        if tstr == "function": return AstNodetype.FUNCTION
        if tstr == "params": return AstNodetype.PARAMS
        if tstr == "variable": return AstNodetype.VARIABLE
        self.error(f"Invalid type {tstr}")



    def parse_parameter_list(self) -> List[AstNodeFunctionParameter]:
        param_list: List[AstNodeFunctionParameter] = []
        self.expect(AstNodetype.LEFT_PAREN)
        while self.current_token.token_type != AstNodetype.RIGHT_PAREN:
            type = AstNodetype.ANY
            if self.is_type():
                type = self.current_token.token_type #self.strToType(self.current_token.value)
                self.advance()
            ident = self.expect(AstNodetype.IDENTIFIER)
            param_list.append(AstNodeFunctionParameter(ident, type))
            if self.current_token.token_type == AstNodetype.COMMA:
                self.advance()
        self.expect(AstNodetype.RIGHT_PAREN)
        return param_list

# expression parser

    def parse_expression(self) -> AstNodeExpression:
        expr = self.parse_logical_or()
        return expr



    def parse_logical_or(self) -> AstNodeExpression:
        node = self.parse_logical_and()
        while self.current_token.token_type == AstNodetype.OR_STATEMENT or self.current_token.token_type == AstNodetype.XOR_STATEMENT:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_logical_and())
        return node



    def parse_logical_and(self) -> AstNodeExpression:
        node = self.parse_equality()
        while self.current_token.token_type == AstNodetype.AND_STATEMENT:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_equality())
        return node



    def parse_equality(self) -> AstNodeExpression:
        node = self.parse_comparison()
        while self.current_token.token_type in [AstNodetype.COMPARETOR_EQUAL, AstNodetype.COMPARETOR_NOT_EQUAL]:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_comparison())
        return node



    def parse_comparison(self) -> AstNodeExpression:
        node = self.parse_addition()
        while self.current_token.token_type in [AstNodetype.COMPARETOR_LESS_THAN, AstNodetype.COMPARETOR_LESS_THAN_EQUAL, AstNodetype.COMPARETOR_GREATER_THAN, AstNodetype.COMPARETOR_GREATER_THAN_EQUAL]:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_addition())
        return node



    def parse_addition(self) -> AstNodeExpression:
        node = self.parse_multiplication()
        while self.current_token.token_type in [AstNodetype.OPERATOR_ADDITION, AstNodetype.OPERATOR_SUBTRACTION]:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_multiplication())
        return node
    


    def parse_multiplication(self) -> AstNodeExpression:
        node = self.parse_power()
        while self.current_token.token_type in [AstNodetype.OPERATOR_MULTIPLICATION, AstNodetype.OPERATOR_DIVISION, AstNodetype.OPERATOR_MODULUS]:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_power())
        return node



    def parse_power(self) -> AstNodeExpression:
        node = self.parse_unary()
        while self.current_token.token_type == AstNodetype.OPERATOR_POWER:
            token = self.current_token
            self.advance()
            node = AstNodeBinaryExpression(token, node, self.parse_unary())
        return node



    def parse_unary(self) -> AstNodeExpression:
        if (self.current_token.token_type == AstNodetype.OPERATOR_SUBTRACTION
                or self.current_token.token_type == AstNodetype.OPERATOR_ADDITION
                or self.current_token.token_type == AstNodetype.NOT_STATEMENT
                or self.current_token.token_type == AstNodetype.BITWISE_NOT):
            token = self.current_token
            self.advance()
            return AstNodeUnaryExpression(token, self.parse_primary())
        return self.parse_primary()



    def parse_primary(self) -> AstNodeExpression:
        if self.current_token.token_type == AstNodetype.LEFT_PAREN:
            self.advance()
            node = self.parse_expression()
            self.expect(AstNodetype.RIGHT_PAREN)
            return node
        if self.current_token.token_type == AstNodetype.IDENTIFIER:
            ident = self.current_token
            self.advance()
            node = None
            if self.current_token.token_type == AstNodetype.LEFT_BRACKET:
                node = AstNodeIndexedVariableExpression(ident)
                while self.current_token.token_type == AstNodetype.LEFT_BRACKET:
                    self.advance()                       # skip left bracket
                    node.index.append(self.parse_expression())
                    self.expect(AstNodetype.RIGHT_BRACKET)
                # TODO: Peek ahead to see if it's a function call.
                if self.current_token == AstNodetype.LEFT_PAREN:
                    self.advance()
                    return AstNodeVariableFunctionCallExpression(node, self.expression_list())
                return node
            if self.current_token.token_type == AstNodetype.LEFT_PAREN:
                self.advance()
                if node is None:
                    node = AstNodeFunctionCallExpression(ident, self.expression_list())
                else:
                    node = AstNodeVariableFunctionCallExpression(node, self.expression_list())
                self.expect(AstNodetype.RIGHT_PAREN)
                return node
            #print(ident)
            #print("self.current_token.token_type", self.current_token.token_type, "self.current_token.value", self.current_token.value, "self.current_token.lineno", self.current_token.lineno, "self.current_token.offset", self.current_token.offset)
            return AstNodeVariableExpression(ident)
        if self.current_token.token_type == AstNodetype.INTEGER:
            token = self.current_token
            self.advance()
            return AstNodeLiteralExpression(token)
        if self.current_token.token_type == AstNodetype.FLOAT:
            token = self.current_token
            self.advance()
            return AstNodeLiteralExpression(token)
        if self.current_token.token_type == AstNodetype.STRING:
            token = self.current_token
            self.advance()
            return AstNodeLiteralExpression(token)
        if self.current_token.token_type == AstNodetype.BOOLEAN_FALSE or self.current_token.token_type == AstNodetype.BOOLEAN_TRUE or self.current_token.token_type == AstNodetype.BOOLEAN:
            token = self.current_token
            self.advance()
            return AstNodeLiteralExpression(token)
        if self.current_token.token_type == AstNodetype.NIL:
            token = self.current_token
            self.advance()
            return AstNodeLiteralExpression(token)
        if self.current_token.token_type == AstNodetype.STRING_INTERPOLATION_START:
            return self.interpolated_string()
        if self.current_token.token_type == AstNodetype.STRING_INTERPOLATION_STRING_PART:
            node = AstNodeLiteralExpression(self.current_token)
            self.advance()
            return node
        if self.current_token.token_type == AstNodetype.NEW_ARRAY or self.current_token.token_type == AstNodetype.ARRAY:
            node = AstNodeLiteralExpression(self.current_token, [], AstNodetype.ARRAY)
            self.advance()
            return node
        if self.current_token.token_type == AstNodetype.NEW_RECORD or self.current_token.token_type == AstNodetype.RECORD:
            node = AstNodeLiteralExpression(self.current_token, {}, AstNodetype.RECORD)
            self.advance()
            return node
        if self.current_token.token_type == AstNodetype.LEFT_BRACKET:
            self.advance()
            node = AstNodeLiteralExpression(self.current_token, self.expression_array(), AstNodetype.ARRAY)
            self.expect(AstNodetype.RIGHT_BRACKET)
            return node
        if self.current_token.token_type == AstNodetype.LEFT_BRACE:
            self.advance()
            node = AstNodeLiteralExpression(self.current_token, self.expression_record(), AstNodetype.RECORD)
            self.expect(AstNodetype.RIGHT_BRACE)
            return node
        if self.current_token.token_type in [AstNodetype.IS_ARRAY_STATEMENT,
                                       AstNodetype.IS_BOOL_STATEMENT,
                                       AstNodetype.IS_FLOAT_STATEMENT,
                                       AstNodetype.IS_FUNCTION_STATEMENT,
                                       AstNodetype.IS_INT_STATEMENT,
                                       AstNodetype.IS_NIL_STATEMENT,
                                       AstNodetype.IS_RECORD_STATEMENT,
                                       AstNodetype.IS_STRING_STATEMENT]:
            return self.is_type_statement()
        if self.current_token.token_type == AstNodetype.TYPE_STATEMENT:
            return self.type_statement()
        self.error(f"Invalid primary expression {self.current_token.token_type} {self.current_token.value}")



    def interpolated_string(self) -> AstNodeExpression:
        token = LexToken.FromToken(self.current_token)
        self.expect(AstNodetype.STRING_INTERPOLATION_START)
        token.token_type = AstNodetype.OPERATOR_ADDITION
        token.value = "+"
        node = AstNodeBinaryExpression(token, AstNodeLiteralExpression(self.current_token), None)
        self.advance() # go to right hand side.
        if node.left._token.token_type == AstNodetype.STRING_INTERPOLATION_END: # if it's just a string without interpolation.
            return node.left
        while self.current_token.token_type != AstNodetype.STRING_INTERPOLATION_END:
            node.right = self.parse_expression()
            node = AstNodeBinaryExpression(token, node, None)
        if self.current_token.token_type == AstNodetype.STRING_INTERPOLATION_END:
            node.right = AstNodeLiteralExpression(self.current_token)
            self.advance()
            return node
        self.error(f"Invalid interpolated string {self.current_token.token_type} {self.current_token.value}")




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This provides the mapping between the bytecode instructions and the mnenomics.
# The mnenomics are in x64 intel notation, that means you first get the destination
# followed by what will be assigned to the destination.
# For programmers this style is intuitive.

#* stack and memory are different places, so we handle certain instructions in special ways.
# Memory is a a Python list that holds tuples of instructions.
# Stack is also a Python list of tuples, but these are type/value pairs.

# The 'CPU' of our VM has the following registers:
# RAX, RBX, RCX, RDS, RSI, RDI, RSP, RBP,
#  R8,  R9,  R10, R11, R12, R13, R14, R15
# RIP, FLAGS

# The stack pointer is RSP.
# The base pointer is RBP.
# The instruction pointer is RIP.
# The flags register is FLAGS.

# The stack pointer is used to point to the top of the stack.
# The base pointer is used to point to the bottom of the stack for a given scope.
# The instruction pointer is used to point to the instruction that is going to be executed,
# it is updated once every execution step.
# The flags register is used to store the flags:
# The flags register in a CPU is used to store the status of the processor after an operation.
# In a Python application representing a virtual CPU,
# the flags register is used to store the status of the processor after an operation.
# The following are the flags and their meanings:
# - **Carry flag (CF)**: Set when there is an overflow from the most significant bit (MSB) of the result. This flag is used to detect unsigned arithmetic overflow.
# - **Zero flag (ZF)**: Set when the result of an operation is zero.
# - **Sign flag (SF)**: Set when the result of an operation is negative.
# - **Overflow flag (OF)**: Set when there is an overflow from the sign bit of the result. This flag is used to detect signed arithmetic overflow.
# - **Direction flag (DF)**: Determines whether string operations increment or decrement their pointers.
# - **Interrupt flag (IF)**: Determines whether interrupts are enabled or disabled.
# - **Trap flag (TF)**: Determines whether single-step debugging is enabled or disabled.
# - **Supervisor flag (SF)**: Determines whether the processor is running in user mode or supervisor mode.
# - **Negative flag (NF)**: Set when the result of an operation has its most significant bit set.

# NOTE: While the individual flags are set or cleared based on the result of an operation,
# the actual value of the flags register itself is not changed.
# so vm.flags.value is always 0, but vm.flags.zero can be True of False.

# NOTE: While the flags register has many flags, only the states for the ZF and CF are maintained.
# The reason for this is plain and simple performance.
# It is very expensive to keep the state of flags, especially the bit state in the register itself.


from typing import Dict, List, Optional

class MacalInstruction:
    def __init__(self,
                 name: str,
                 opcode: int,
                 operands: Optional[List[str]] = None,
                 description: Optional[str] = None):
        self.Name: str = name
        self.Opcode: int = opcode
        self.Operands: List[str] = [] if operands is None else operands
        self.Description: str = "" if description is None else description

    def __str__(self) -> str:
        return f"{self.Name} {', '.join(self.Operands)}"

    def __repr__(self) -> str:
        return self.__str__()

class MacalInstructionList:
    def __init__(self):
        self.instructions: Dict[int, MacalInstruction] = {}
        self.named_instructions: Dict[str, MacalInstruction] = {}
        # These are the instructions that are implemented in the VM.
        self.add("NOP",      [],                "no operation")
        self.add("JMP",      ["addr"],          "jump to addr")
        self.add("JMPZ",     ["addr"],          "jump to addr if zero flag set")
        self.add("JMPNZ",    ["addr"],          "jump to addr if zero flag not set")
        self.add("JMPLT",    ["addr"],          "jump to addr if less than flag set")
        self.add("JMPGT",    ["addr"],          "jump to addr if greater than flag set")
        self.add("JMPLTE",   ["addr"],          "jump to addr if less than or equal flag set")
        self.add("JMPGTE",   ["addr"],          "jump to addr if greater than or equal flag set")
        self.add("DJNZ",     ["addr"],          "decrement rcx, jump to addr if not zero for loops.")
        self.add("PUSH",     ["reg"],           "put register on stack")
        self.add("POP",      ["reg"],           "get register from stack")
        self.add("MOVRR",    ["reg1", "reg2"],  "reg1 = reg2")
        self.add("MOVRI",    ["reg", "imm"],    "reg  = imm")
        self.add("MOVMI",    ["[addr]", "imm"], "sets the value at addr to imm")
        self.add("MOVRM",    ["reg", "[addr]"], "reg  = value at addr")
        self.add("MOVMR",    ["[addr]", "reg"], "value at addr = reg")
        self.add("CMPRI",    ["reg", "imm"],    "compare value of reg to imm, sets ZF, SF, CF, OF")
        self.add("CMPRR",    ["reg1", "reg2"],  "compare value of reg1 to reg2, sets ZF, SF, CF, OF")
        self.add("ADDRI",    ["reg", "imm"],    "add imm to reg")
        self.add("ADDRR",    ["reg1", "reg2"],  "add reg2 to reg1")
        self.add("SUBRI",    ["reg", "imm"],    "subtract imm from reg")
        self.add("SUBRR",    ["reg1", "reg2"],  "subtract reg2 from reg1")
        self.add("MULRI",    ["reg", "imm"],    "multiply reg by imm")
        self.add("MULRR",    ["reg1", "reg2"],  "multiply reg1 by reg2")
        self.add("DIVRI",    ["reg", "imm"],    "divide reg by imm")
        self.add("DIVRR",    ["reg1", "reg2"],  "divide reg1 by reg2")
        self.add("MODRI",    ["reg", "imm"],    "modulus reg by imm")
        self.add("MODRR",    ["reg1", "reg2"],  "modulus reg1 by reg2")
        self.add("POWRR",    ["reg1", "reg2"],  "reg1 = reg1 ^ reg2")
        self.add("POWRI",    ["reg", "imm"],    "reg = reg ^ imm")
        self.add("ANDRI",    ["reg", "imm"],    "and reg by imm")
        self.add("ANDRR",    ["reg1", "reg2"],  "and reg1 by reg2")
        self.add("ORRI",     ["reg", "imm"],    "or reg by imm")
        self.add("ORRR",     ["reg1", "reg2"],  "or reg1 by reg2")
        self.add("XORI",     ["reg", "imm"],    "xor reg by imm")
        self.add("XOR",      ["reg1", "reg2"],  "xor reg1 by reg2, sets ZF if result is zero")
        self.add("NEG",      ["reg"],           "negate reg")
        self.add("NOT",      ["reg"],           "not reg")
        self.add("SETZR",    ["reg"],           "set reg.value = 1 if ZF = 1")
        self.add("SETNZR",   ["reg"],           "set reg.value = 1 if ZF = 0")
        self.add("SETLR",    ["reg"],           "set reg.value = 1 if CF = 1 and ZF = 0")
        self.add("SETGR",    ["reg"],           "set reg.value = 1 if CF = 0 and ZF = 0")
        self.add("SETLER",   ["reg"],           "set reg.value = 1 if CF = 1 or ZF = 1")
        self.add("SETGER",   ["reg"],           "set reg.value = 1 if CF = 0 or ZF = 1")
        self.add("CALL",     ["addr"],          "call function at addr")
        self.add("CALLR",    ["reg"],           "call function at reg")
        self.add("CALLE",    ["module", "function", "paramcount"], "call function in external module.")
        self.add("RET",      [],                "return from function")
        self.add("RETZ",     [],                "return from function if zero flag set")
        self.add("RETNZ",    [],                "return from function if zero flag not set")
        self.add("HALT",     [],                "halt execution")
        self.add("INC",      ["reg"],           "reg = reg + 1")
        self.add("DEC",      ["reg"],           "reg = reg - 1")
        self.add("LEN",      ["reg"],           "reg = len(reg)")
        self.add("LENR",     ["reg1", "reg2"],  "reg1 = len(reg2)")
        self.add("INKEYS",   ["reg1", "reg2"],  "reg1 = reg1 in reg2.keys()")
        self.add("LOAD",     ["reg", "offset"], "Loads a variable from stack offset into reg")
        self.add("STOR",     ["offset", "reg"], "Stores a variable from reg into stack offset")
        self.add("LOADR",    ["reg", "reg", "offset"], "Loads a variable from stack offset relative to rbp into reg")
        self.add("STORR",    ["reg", "offset", "reg"], "Stores a variable from reg into stack offset, offset is relative to rbp")
        self.add("STORIR",   ["reg1", "reg2"],  "Stores a variable from reg2 into stack at the offset in reg1")
        self.add("CMPRTI",   ["reg", "type"],    "compare value type information of reg to imm, sets ZF, SF, CF, OF")
        self.add("SETRTI" ,  ["reg", "type"],    "set reg.type = imm")
        self.add("APPEND" ,  ["reg1", "reg2"],  "append reg1[] = reg2")
        self.add("ERROR",    ["errmsg"],        "print errmsg and halt")
        self.add("ERRORZ",   ["errmsg"],        "if zero flag set, print errmsg and halt")
        self.add("ERRORNZ",  ["errmsg"],        "if zero flag is not set, print errmsg and halt")
        self.add("PRNT",     [],                "print rax")
        self.add("INDEX",    [],                "rax = rax[rbx]")
        self.add("FEINDEX",  [],                "rax = rax[rbx] with the special note that rbx is the index to the keys() of a record if rax is a record, otherwise just like INDEX")
        self.add("INDEXR",   ["reg"],           "rax[rbx] = reg")
        self.add("HASFLDRR", ["reg1", "reg2"],  "reg1 = reg1.has_field(reg2)")
        self.add("MOVRTOA",  ["reg"],           "reg = [reg], the record in reg is moved into an array with that record as an element.")
        self.add("TYPE",     ["reg"],           "reg = type(reg)")
        # These instructions are for debugging purposes only.
        self.add("STACK",    [],                "show stack for debugging")
        self.add("REGS",     [],                "show registers for debugging")
        self.add("PRNTR",    ["reg"],           "print reg, debug printing register name, value, type")
        self.add("DATASEG",  ["metadata"],      "set metadata for debugging")
        self.add("MERGE",    ["reg1", "reg2", "reg3"],  "merge reg2 and reg3 into reg1 reg1 = [reg2] + [reg3]")
        self.add("JMPR",     ["reg"],           "jump to reg")
        self.add("RESERVE",  ["imm"],           "reserve imm entries on the stack")
        self.add("RESERVEDS", ["metadata"],     "datasegment with metadata for variables that are reserved on the stack")
        self.add("LOADIR",   ["reg1", "reg2"],  "reg1 = stack[reg2]")
        self.add("MOVATOR",  ["reg"],           "reg = reg[0], the first record in the array in reg is moved into reg.")
        self.add("DEBUGSTR", ["str"],           "print str for debugging")
        self.add("MOVSR",    ["reg"],           "reg = stack[rsp-1]")

    def add(self, name: str, operands: Optional[List[str]] = None, description: Optional[str] = None):
        opcode = len(self.instructions)
        instruction = MacalInstruction(name, opcode, operands, description)
        self.instructions[opcode] = instruction
        self.named_instructions[name] = instruction

    def get(self, opcode: int) -> MacalInstruction:
        if opcode not in self.instructions:
            raise ValueError(f"Unknown opcode {opcode}")
        return self.instructions[opcode]

    def getByName(self, name: str) -> MacalInstruction:
        if name not in self.named_instructions:
            raise ValueError(f"Unknown instruction {name}")
        return self.named_instructions[name]

    def __getitem__(self, opcode: int) -> MacalInstruction:
        return self.get(opcode)

    def __getattr__(self, name: str) -> int:
        return self.getByName(name).Opcode

    def __len__(self) -> int:
        return len(self.instructions)

    def __iter__(self):
        return iter(self.instructions.values())

    def __str__(self) -> str:
        return "\n".join([str(instruction) for instruction in self.named_instructions.values()])

    def __repr__(self) -> str:
        return self.__str__()

    def fromOpcode(self, opcode: int) -> MacalInstruction:
        return self.get(opcode)

Opcode = MacalInstructionList()